@@ -6,6 +6,10 @@ use strict;
 use Carp;
 use Config;
 use Scalar::Util qw/blessed/;
+use Class::Method::Modifiers;
+use Package::Stash;
+use Try::Tiny;
+use namespace::clean;   # important: clean all subs imported above this line
 
 =head1 NAME
 
@@ -13,11 +17,11 @@ Test::Mock::Redis - use in place of Redis for unit testing
 
 =head1 VERSION
 
-Version 0.07
+Version 0.14
 
 =cut
 
-our $VERSION = '0.07';
+our $VERSION = '0.14';
 
 =head1 SYNOPSIS
 
@@ -38,7 +42,14 @@ This module is designed to function as a drop in replacement for
 Redis.pm for testing purposes.
 
 See perldoc Redis and the redis documentation at L<http://redis.io>
-    
+
+=head1 PERSISTENCE
+
+The "connection" to the mocked server (and its stored data) will persist beyond
+the object instance, just like a real Redis server. This means that you do not
+need to save the instance to this object in order to preserve your data; simply
+call C<new> with the same server parameter and the same instance will be
+returned, with all data preserved.
 
 =head1 SUBROUTINES/METHODS
 
@@ -50,23 +61,51 @@ See perldoc Redis and the redis documentation at L<http://redis.io>
 
     It accepts the "server" argument, just like Redis.pm's new.
 
-=cut
+=head2 num_databases
 
-sub _new_db {
-    tie my %hash, 'Test::Mock::Redis::PossiblyVolitile'; 
-    return \%hash;
-}
+Redis ships with a default of 16 databases, and that's what this module 
+handles by default. If you need to change that, do
+
+    use Test::Mock::Redis num_databases => 21;
+
+or at run-time
+
+    Test::Mock::Redis::change_num_databases(21);
+
+=cut
 
 my $NUM_DBS = 16;
 
+sub import {
+    my ($class, %args) = @_;
+
+    if ($args{num_databases}){
+        change_num_databases($args{num_databases});
+    }
+}
+
+sub change_num_databases {
+    $NUM_DBS = shift;
+}
+
+
+sub _new_db {
+    tie my %hash, 'Test::Mock::Redis::PossiblyVolatile';
+    return \%hash;
+}
+
+
 sub _defaults {
+    my @hex = (0..9, 'a'..'f');
     return (
         _quit      => 0,
         _shutdown  => 0,
         _stash     => [ map { _new_db } (1..$NUM_DBS) ],
+        _num_dbs   => $NUM_DBS,
         _db_index  => 0,
         _up_since  => time,
         _last_save => time,
+        _run_id    => (join '', map { $hex[rand @hex] } 1..40), # E.G. '0e7e19fc45139fdb26ff3dd35ca6725d9882f1b7',
     );
 }
 
@@ -106,14 +145,16 @@ sub auth {
 
     confess '[auth] ERR wrong number of arguments for \'auth\' command' unless @_;
 
-    return 1;
+    return 'OK';
 }
 
 sub quit {
     my $self = shift;
 
-    confess "Not connected to any server" if $self->{_quit};
+    my $return = !$self->{_quit};
+
     $self->{_quit} = 1;
+    return $return;
 }
 
 sub shutdown {
@@ -126,7 +167,7 @@ sub set {
     my ( $self, $key, $value ) = @_;
 
     $self->_stash->{$key} = "$value";
-    return 1;
+    return 'OK';
 }
 
 sub setnx {
@@ -143,7 +184,7 @@ sub setex {
     my ( $self, $key, $ttl, $value ) = @_;
     $self->set($key, $value);
     $self->expire($key, $ttl);
-    return 1;
+    return 'OK';
 }
 
 sub expire {
@@ -265,7 +306,7 @@ sub mset {
 
     @{ $self->_stash }{keys %things} = (values %things);
 
-    return 1;
+    return 'OK';
 }
 
 sub msetnx {
@@ -292,7 +333,7 @@ sub type {
     my ( $self, $key ) = @_;
     # types are string, list, set, zset and hash
 
-    return undef unless $self->exists($key);
+    return 'none' unless $self->exists($key);
 
     my $type = ref $self->_stash->{$key};
 
@@ -319,7 +360,10 @@ sub keys :method {
     $match =~ s/(?<!\\)\*/.*/g;
     $match =~ s/(?<!\\)\?/.?/g;
 
-    return @{[ sort { $a cmp $b } grep { /$match/ } keys %{ $self->_stash }]};
+    return @{[ sort { $a cmp $b }
+        grep { exists $self->_stash->{$_} }
+        grep { /$match/ }
+        keys %{ $self->_stash }]};
 }
 
 sub randomkey {
@@ -340,7 +384,7 @@ sub rename {
 
     $self->_stash->{$to} = $self->_stash->{$from};
     delete $self->_stash->{$from};
-    return 1;
+    return 'OK';
 }
 
 sub renamenx {
@@ -361,7 +405,8 @@ sub rpush {
 
     $self->_make_list($key);
 
-    return push @{ $self->_stash->{$key} }, "$value";
+    push @{ $self->_stash->{$key} }, "$value";
+    return scalar @{ $self->_stash->{$key} };
 }
 
 sub lpush {
@@ -372,7 +417,8 @@ sub lpush {
 
     $self->_make_list($key);
 
-    return unshift @{ $self->_stash->{$key} }, "$value";
+    unshift @{ $self->_stash->{$key} }, "$value";
+    return scalar @{ $self->_stash->{$key} };
 }
 
 sub rpushx {
@@ -380,7 +426,8 @@ sub rpushx {
 
     return unless $self->_is_list($key);
 
-    return push @{ $self->_stash->{$key} }, "$value";
+    push @{ $self->_stash->{$key} }, "$value";
+    return scalar @{ $self->_stash->{$key} };
 }
 
 sub lpushx {
@@ -388,7 +435,8 @@ sub lpushx {
 
     return unless $self->_is_list($key);
 
-    return unshift @{ $self->_stash->{$key} }, "$value";
+    unshift @{ $self->_stash->{$key} }, "$value";
+    return scalar @{ $self->_stash->{$key} };
 }
 
 sub llen {
@@ -409,7 +457,7 @@ sub ltrim {
     my ( $self, $key, $start, $end ) = @_;
 
     $self->_stash->{$key} = [ @{ $self->_stash->{$key} }[$start..$end] ]; 
-    return 1;
+    return 'OK';
 }
 
 sub lindex {
@@ -422,7 +470,7 @@ sub lset {
     my ( $self, $key, $index, $value ) = @_;
 
     $self->_stash->{$key}->[$index] = "$value";
-    return 1;
+    return 'OK';
 }
 
 sub lrem {
@@ -463,8 +511,13 @@ sub rpop {
 sub select {
     my ( $self, $index ) = @_;
 
+    my $max_index = $#{ $self->{_stash} };
+    if ($index > $max_index ){
+        die "You called select($index), but max allowed is $max_index unless you configure more databases"; 
+    }
+
     $self->{_db_index} = $index;
-    return 1;
+    return 'OK';
 }
 
 sub _stash {
@@ -789,7 +842,7 @@ sub sort {
 sub save { 
     my $self = shift;
     $self->{_last_save} = time;
-    return 1;
+    return 'OK';
 }
 
 sub bgsave { 
@@ -806,33 +859,62 @@ sub info {
     my $self = shift;
 
     return {
-        arch_bits                  => $Config{use64bitint } ? '64' : '32',
-        bgrewriteaof_in_progress   => '0',
-        bgsave_in_progress         => '0',
-        blocked_clients            => '0',
-        changes_since_last_save    => '0',
-        connected_clients          => '1',
-        connected_slaves           => '0',
-        expired_keys               => '0',
-        hash_max_zipmap_entries    => '64',
-        hash_max_zipmap_value      => '512',
-        last_save_time             => $self->{_last_save},
-        mem_fragmentation_ratio    => '0.11',
-        multiplexing_api           => 'kqueue',
-        process_id                 => $$,
-        pubsub_channels            => '0',
-        pubsub_patterns            => '0',
-        redis_git_dirty            => '0',
-        redis_git_sha1             => 'da14590b',
-        redis_version              => '2.1.4',
-        role                       => 'master',
-        total_commands_processed   => '84',
-        total_connections_received => '14',
+        aof_current_rewrite_time_sec => '-1',
+        aof_enabled => '0',
+        aof_last_bgrewrite_status => 'ok',
+        aof_last_rewrite_time_sec => '-1',
+        aof_rewrite_in_progress => '0',
+        aof_rewrite_scheduled => '0',
+        arch_bits => $Config{use64bitint } ? '64' : '32',
+        blocked_clients => '0',
+        client_biggest_input_buf => '0',
+        client_longest_output_list => '0',
+        connected_clients => '1',
+        connected_slaves => '0',
+        evicted_keys => '0',
+        expired_keys => '0',
+        gcc_version => '4.2.1',
+        instantaneous_ops_per_sec => '568',
+        keyspace_hits => '272',
+        keyspace_misses => '0',
+        latest_fork_usec => '0',
+        loading => '0',
+        lru_clock => '1994309',
+        mem_allocator => 'libc',
+        mem_fragmentation_ratio => '1.61',
+        multiplexing_api => 'kqueue',
+        os => $Config{osname}.' '.$Config{osvers},  # should be like 'Darwin 12.2.1 x86_64', this is close
+        process_id => $$,
+        pubsub_channels => '0',
+        pubsub_patterns => '0',
+        rdb_bgsave_in_progress => '0',
+        rdb_changes_since_last_save => '0',
+        rdb_current_bgsave_time_sec => '-1',
+        rdb_last_bgsave_status => 'ok',
+        rdb_last_bgsave_time_sec => '-1',
+        rdb_last_save_time => '1362120372',
+        redis_git_dirty => '0',
+        redis_git_sha1 => '34b420db',
+        redis_mode => 'standalone',
+        redis_version => '2.6.10',
+        rejected_connections => '0',
+        role => 'master',
+        run_id => $self->{_run_id},
+        tcp_port => '11084',
+        total_commands_processed => '1401',
+        total_connections_received => '1',
         uptime_in_days             => (time - $self->{_up_since}) / 60 / 60 / 24,
         uptime_in_seconds          => time - $self->{_up_since},
-        used_memory                => '3918288',
-        used_memory_human          => '3.74M',
-        vm_enabled                 => '0',
+        used_cpu_sys => '0.04',
+        used_cpu_sys_children => '0.00',
+        used_cpu_user => '0.02',
+        used_cpu_user_children => '0.00',
+        used_memory => '1056288',
+        used_memory_human => '1.01M',
+        used_memory_lua => '31744',
+        used_memory_peak => '1055728',
+        used_memory_peak_human => '1.01M',
+        used_memory_rss => '1699840',
         map { 'db'.$_ => sprintf('keys=%d,expires=%d',
                              scalar keys %{ $self->_stash($_) },
                              $self->_expires_count_for_db($_),
@@ -971,6 +1053,13 @@ sub zremrangebyscore {
     return scalar @remove;
 }
 
+=head1 PIPELINING
+
+See L<Redis/PIPELINING> -- most methods support the use of a callback sub as
+the final argument. For this implementation, the callback sub will be called
+immediately (before the result of the original method is returned), and
+C<wait_all_responses> does nothing.  Combining pipelining with C<multi>/C<exec>
+is not supported.
 
 =head1 TODO
 
@@ -1049,15 +1138,25 @@ Salvatore Sanfilippo for redis, of course!
 
 Dobrica Pavlinusic & Pedro Melo for Redis.pm
 
+The following people have contributed to I<Test::Mock::Redis>:
+
+=over
+
+=item * Karen Etheridge
+
+=item * Kevin Goess
+
+=back
+
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 2011 Jeff Lavallee.
+Copyright 2011, 2012, 2013 Jeff Lavallee.
 
 This program is free software; you can redistribute it and/or modify it
 under the terms of either: the GNU General Public License as published
 by the Free Software Foundation; or the Artistic License.
 
-See http://dev.perl.org/licenses/ for more information.
+See L<http://dev.perl.org/licenses/> for more information.
 
 
 =cut
@@ -1124,6 +1223,172 @@ sub _make_zset {
 }
 
 
+# MULTI/EXEC/DISCARD: http://redis.io/topics/transactions
+
+sub multi {
+    my ( $self ) = @_;
+
+    confess '[multi] ERR MULTI calls can not be nested' if defined $self->{_multi_commands};
+
+    # set up the list for storing commands sent between MULTI and EXEC/DISCARD
+    $self->{_multi_commands} = [];
+
+    return 'OK';
+}
+
+# methods that return a list, rather than a single value
+my @want_list = qw(mget keys lrange smembers sinter sunion sdiff hmget hkeys hvals hgetall sort zrange zrevrange zrangebyscore);
+my %want_list = map { $_ => 1 } @want_list;
+
+sub exec {
+    my ( $self ) = @_;
+
+    # we are going to commit all the changes we saved up;
+    # replay them now and return all their output
+
+    confess '[exec] ERR EXEC without MULTI' if not defined $self->{_multi_commands};
+
+    my @commands = @{$self->{_multi_commands}};
+    delete $self->{_multi_commands};
+
+    # replay all the queries that were queued up
+    # the returned result is a nested array of the results of all the commands
+    my @exceptions;
+    my @results = map {
+        my ($method, @args) = @$_;
+        my @result =
+            try { $self->$method(@args) }
+            catch { push @exceptions, $_; (); };
+        $want_list{$method} ? \@result : $result[0];
+    } @commands;
+
+    s/^\[\w+\] // for @exceptions;
+
+    confess('[exec] ', join('; ', @exceptions)) if @exceptions;
+
+    return @results;
+}
+
+sub discard {
+    my ( $self ) = @_;
+
+    confess '[discard] ERR DISCARD without MULTI' if not defined $self->{_multi_commands};
+
+    # discard all the accumulated commands, without executing them
+    delete $self->{_multi_commands};
+
+    return 'OK';
+}
+
+# now that we've defined all our subs, we need to wrap them all in logic that
+# can check if we are in the middle of a MULTI, and if so, queue up the
+# commands for later replaying.
+
+my %no_transaction_wrap_methods = (
+    new => 1,
+    multi => 1,
+    exec => 1,
+    discard => 1,
+    quit => 1,
+    import => 1,
+    change_num_databases => 1,
+);
+
+my @transaction_wrapped_methods =
+    grep { !/^_/}
+    grep { not $no_transaction_wrap_methods{$_} }
+        Package::Stash->new(__PACKAGE__)->list_all_symbols('CODE');
+
+foreach my $method (@transaction_wrapped_methods)
+{
+    around $method => sub {
+        my $orig = shift;
+        my $self = shift;
+
+        # pass command through if we are not handling a MULTI
+        return $self->$orig(@_) if not defined $self->{_multi_commands};
+
+        push @{$self->{_multi_commands}}, [ $method, @_ ];
+        return 'QUEUED';
+    };
+}
+
+
+# PIPELINING SUPPORT
+
+# these method modifications must be done after (over top of) the modification
+# for transactions, as we need to check for/extract the $cb first.
+
+my %no_pipeline_wrap_methods = (
+    new => 1,
+    multi => 1,
+    discard => 1,
+    quit => 1,
+    ping => 1,
+    subscribe => 1,
+    unsubscribe => 1,
+    psubscribe => 1,
+    punsubscribe => 1,
+    wait_all_responses => 1,
+);
+
+my @pipeline_wrapped_methods =
+    grep { !/^_/}
+    grep { not $no_pipeline_wrap_methods{$_} }
+        Package::Stash->new(__PACKAGE__)->list_all_symbols('CODE');
+
+# this is a bit messy, and the wantarray logic may not be quite right.
+# Alternatively, we could implement all this by reusing the logic in the real
+# Redis.pm -- subclass Redis, override new/multi/exec/discard (and probably
+# some other special functions), and have __run_cmd use a dispatch table to
+# call all our overridden implementations.
+
+foreach my $method (@pipeline_wrapped_methods)
+{
+    around $method => sub {
+        my $orig = shift;
+        my $self = shift;
+        my @args = @_;
+
+        my $cb = @args && ref $args[-1] eq 'CODE' ? pop @args : undef;
+
+        return $self->$orig(@args) if not $cb;
+
+        # this may be officially supported eventually -- see
+        # https://github.com/melo/perl-redis/issues/17
+        # and "Pipeline management" in the Redis docs
+        # To make this work, we just need to special-case exec, to collect all the
+        # results and errors in tuples and send that to the $cb
+        die 'cannot combine pipelining with MULTI' if $self->{_multi_commands};
+
+        # We could also implement this with a queue, not bothering to process
+        # the commands until wait_all_responses is called - but then we need to
+        # make sure to call wait_all_responses explicitly as soon as a command
+        # is issued without a $cb.
+
+        my $error;
+        my (@result) = try
+        {
+            $self->$orig(@args);
+        }
+        catch
+        {
+            $error = $_;
+            ();
+        };
+
+        $cb->(
+            # see notes above - this logic may not be quite right
+            ( $want_list{$method} ? \@result : $result[0] ),
+            $error,
+        );
+        return 1;
+    };
+}
+
+# in a real Redis system, this will make all outstanding callbacks get called.
+sub wait_all_responses {}
+
 
 1; # End of Test::Mock::Redis
 
@@ -1143,7 +1408,7 @@ package Test::Mock::Redis::Set;
 sub new { return bless {}, shift }
 1;
 
-package Test::Mock::Redis::PossiblyVolitile;
+package Test::Mock::Redis::PossiblyVolatile;
 
 use strict; use warnings;
 use Tie::Hash;
